/*
*   quicktime_codec_divx plugin for quicktime4linux library
*   enables decoding and encoding with DivX
*
*   (c) 27.8.2001 Ingo D. Rullhusen <d01c@uni-bremen.de>
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or   
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of 
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*   see gpl.txt for full license
*
*   ANNOTATION: DivX comes with an own license!
*/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <openquicktime.h>
#include <colormodels.h>


#include <encore2.h>
#include <decore.h>


typedef struct
{
   ENC_PARAM enc_param;
   
   unsigned long dec_handle;
   DEC_PARAM dec_param;
   DEC_MEM_REQS dec_mem_reqs;
   DEC_SET dec_set;
   
   unsigned char *buffer;
   int quality;
   int dec_cmodel;
} CODEC_PARAM;


static void RGBswitch( CODEC_PARAM *codec_param,
                       unsigned char *in, unsigned char *out )
{
   int i, j;
   unsigned char *ic, *oc;

   oc = out;
   for ( i=codec_param->enc_param.y_dim-1; i>-1; i-- )
   {
      ic = in + i*3*codec_param->enc_param.x_dim;
      for ( j=0; j<codec_param->enc_param.x_dim; j++ )
      {
         *oc = *(ic+2);
         oc++;
         *oc = *(ic+1);
         oc++;
         *oc = *ic;
         oc++; ic+=3;
      }
   }

}


static void RGB32switch( CODEC_PARAM *codec_param,
                         unsigned char *in, unsigned char *out )
{
   int i, j;
   unsigned char *ic, *oc;

   oc = out;
   for ( i=codec_param->enc_param.y_dim-1; i>-1; i-- )
   {
      ic = in + i*4*codec_param->enc_param.x_dim;
      for ( j=0; j<codec_param->enc_param.x_dim; j++ )
      {
         *oc = *(ic+2);
         oc++;
         *oc = *(ic+1);
         oc++;
         *oc = *ic;
         oc++;
         *oc = *(ic+1);
         oc++; ic+=4;
      }
   }

}


static int reinit_divx(CODEC_PARAM *codec_param)
{
   encore(codec_param->enc_param.handle, ENC_OPT_RELEASE, &codec_param->enc_param, NULL);
   return( encore(NULL, ENC_OPT_INIT, &codec_param->enc_param, NULL) );
}

int set_param_divx( quicktime_t* file, int track,
                    const char *param, const void *data )
{
   CODEC_PARAM *p = (CODEC_PARAM*)(((quicktime_codec_t*)(file->vtracks[track].codec))->priv);

   if (!strcmp(param,"bitrate"))
   {
      p->enc_param.bitrate = *((long*)data);
   }
   else
   if (!strcmp(param,"rc_period"))
   {
      p->enc_param.rc_period = *((int*)data);
   }
   else
   if (!strcmp(param,"rc_reaction_period"))
   {
      p->enc_param.rc_reaction_period = *((int*)data);
   }
   else
   if (!strcmp(param,"rc_reaction_ratio"))
   {
      p->enc_param.rc_reaction_ratio = *((int*)data);
   }
   else
   if (!strcmp(param,"max_quantizer"))
   {
      p->enc_param.max_quantizer = *((int*)data);
   }
   else
   if (!strcmp(param,"min_quantizer"))
   {
      p->enc_param.min_quantizer = *((int*)data);
   }
   else
   if (!strcmp(param,"max_key_interval"))
   {
      p->enc_param.max_key_interval = *((int*)data);
   }
   else
   if (!strcmp(param,"use_bidirect"))
   {
      p->enc_param.use_bidirect = *((int*)data);
   }
   else
   if (!strcmp(param,"deinterlace"))
   {
      p->enc_param.deinterlace = *((int*)data);
   }
   else
   if (!strcmp(param,"quality"))
   {
      p->enc_param.quality = *((int*)data);
   }
   else
   if (!strcmp(param,"obmc"))
   {
      p->enc_param.obmc = *((int*)data);
   }
   else
   {
      fprintf(stderr,"set_param_divx: unknown parameter named '%s'\n",param);
      return(1);
   }

   if ( reinit_divx(p) == ENC_OK )
   {
      return(0);
   }
   
   fprintf(stderr,"set_param_divx: reinit failed\n");
   return(1);
}


int get_param_divx(quicktime_t *file, int track, const char *param, void *data)
{
   CODEC_PARAM *p = (CODEC_PARAM*)(((quicktime_codec_t*)(file->vtracks[track].codec))->priv);

   if (!strcmp(param,"bitrate"))
   {
      *((long*)data) = p->enc_param.bitrate;
      return(QUICKTIME_UNKNOWN_PARAMETER);
   }
   else
   if (!strcmp(param,"rc_period"))
   {
      *((int*)data) = p->enc_param.rc_period;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"rc_reaction_period"))
   {
      *((int*)data) = p->enc_param.rc_reaction_period;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"rc_reaction_ratio"))
   {
      *((int*)data) = p->enc_param.rc_reaction_ratio;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"max_quantizer"))
   {
      *((int*)data) = p->enc_param.max_quantizer;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"min_quantizer"))
   {
      *((int*)data) = p->enc_param.min_quantizer;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"max_key_interval"))
   {
      *((int*)data) = p->enc_param.max_key_interval;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"use_bidirect"))
   {
      *((int*)data) = p->enc_param.use_bidirect;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"deinterlace"))
   {
      *((int*)data) = p->enc_param.deinterlace;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"quality"))
   {
      *((int*)data) = p->enc_param.quality;
      return(QUICKTIME_INTEGER_PARAMETER);
   }
   else
   if (!strcmp(param,"obmc"))
   {
      *((int*)data) = p->enc_param.obmc;
      return(QUICKTIME_INTEGER_PARAMETER);
   }

   fprintf(stderr,"get_param_divx: unknown parameter named '%s'\n",param);
   return(1);
}


void *init_divx( quicktime_video_map_t *qvm )
{
   CODEC_PARAM *codec_param;   

   codec_param = calloc( 1, sizeof(CODEC_PARAM) );

   codec_param->enc_param.x_dim = (int)rint(qvm->track->tkhd.track_width);
   codec_param->enc_param.y_dim = (int)rint(qvm->track->tkhd.track_height);
   codec_param->enc_param.framerate = (float)qvm->track->mdia.mdhd.time_scale /
                                      (float)qvm->track->mdia.minf.stbl.stts.table[0].sample_duration;

   codec_param->buffer = malloc( 3*codec_param->enc_param.x_dim
                                  *codec_param->enc_param.y_dim );

   if ( 0 > qvm->track->mdia.mdhd.quality || qvm->track->mdia.mdhd.quality > 100 )
      qvm->track->mdia.mdhd.quality = 100;

   codec_param->quality = qvm->track->mdia.mdhd.quality;
   codec_param->enc_param.rc_period = 2000;
   codec_param->enc_param.rc_reaction_period = 10;
   codec_param->enc_param.rc_reaction_ratio = 20;
   codec_param->enc_param.max_key_interval = (int)rint(3*codec_param->enc_param.framerate);
   codec_param->enc_param.use_bidirect = 0;
   codec_param->enc_param.deinterlace = 0;
   codec_param->enc_param.obmc = 0;

   codec_param->enc_param.bitrate = (long)rint( 8*3*codec_param->enc_param.x_dim*codec_param->enc_param.y_dim
                                                *codec_param->enc_param.framerate
                                                * 0.03 * codec_param->quality/100.0 );

   codec_param->enc_param.min_quantizer = 12 - codec_param->quality/10;
   codec_param->enc_param.max_quantizer = codec_param->enc_param.min_quantizer + 2;

   codec_param->enc_param.quality = 1 + codec_param->quality/25;

   encore(NULL, ENC_OPT_INIT, &codec_param->enc_param, NULL);

   codec_param->dec_param.x_dim = (int)rint(qvm->track->tkhd.track_width);
   codec_param->dec_param.y_dim = (int)rint(qvm->track->tkhd.track_height);

   codec_param->dec_param.output_format = DEC_RGB32;
   codec_param->dec_cmodel = BC_RGBA8888;

   codec_param->dec_param.time_incr = 15;

   codec_param->dec_handle = (long unsigned)codec_param->enc_param.handle + 1;

   decore(codec_param->dec_handle, DEC_OPT_MEMORY_REQS, &codec_param->dec_param, &codec_param->dec_mem_reqs);

   codec_param->dec_param.buffers.mp4_edged_ref_buffers =
      malloc(codec_param->dec_mem_reqs.mp4_edged_ref_buffers_size);
   codec_param->dec_param.buffers.mp4_edged_for_buffers =
      malloc(codec_param->dec_mem_reqs.mp4_edged_for_buffers_size);
   codec_param->dec_param.buffers.mp4_edged_back_buffers =
      malloc(codec_param->dec_mem_reqs.mp4_edged_back_buffers_size);
   codec_param->dec_param.buffers.mp4_display_buffers =
      malloc(codec_param->dec_mem_reqs.mp4_display_buffers_size);
   codec_param->dec_param.buffers.mp4_state =
      malloc(codec_param->dec_mem_reqs.mp4_state_size);
   codec_param->dec_param.buffers.mp4_tables =
      malloc(codec_param->dec_mem_reqs.mp4_tables_size);
   codec_param->dec_param.buffers.mp4_stream =
      malloc(codec_param->dec_mem_reqs.mp4_stream_size);
   codec_param->dec_param.buffers.mp4_reference =
      malloc(codec_param->dec_mem_reqs.mp4_reference_size);

   memset(codec_param->dec_param.buffers.mp4_state, 0, codec_param->dec_mem_reqs.mp4_state_size);
   memset(codec_param->dec_param.buffers.mp4_tables, 0, codec_param->dec_mem_reqs.mp4_tables_size);
   memset(codec_param->dec_param.buffers.mp4_stream, 0, codec_param->dec_mem_reqs.mp4_stream_size);
   memset(codec_param->dec_param.buffers.mp4_reference, 0, codec_param->dec_mem_reqs.mp4_reference_size);

   decore(codec_param->dec_handle, DEC_OPT_INIT, &codec_param->dec_param, NULL);

   codec_param->dec_set.postproc_level = 0;
   
   decore(codec_param->dec_handle, DEC_OPT_SETPP, &codec_param->dec_set, NULL);

   return(codec_param);
}


static int change_dec_cmodel( CODEC_PARAM *codec_param, int cmodel )
{
   switch( cmodel )
   {
      case BC_RGB888:
         codec_param->dec_cmodel = BC_RGB888;
         codec_param->dec_param.output_format = DEC_RGB24;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      case BC_RGBA8888:
         codec_param->dec_cmodel = BC_RGBA8888;
         codec_param->dec_param.output_format = DEC_RGB32;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      case BC_BGR888:
         codec_param->dec_cmodel = BC_BGR888;
         codec_param->dec_param.output_format = DEC_RGB24_INV;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      case BC_BGRA8888:
         codec_param->dec_cmodel = BC_BGRA8888;
         codec_param->dec_param.output_format = DEC_RGB32_INV;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      case BC_YUV420P:
         codec_param->dec_cmodel = BC_YUV420P;
         codec_param->dec_param.output_format = DEC_YV12;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      case BC_BGR565:
         codec_param->dec_cmodel = BC_BGR565;
         codec_param->dec_param.output_format = DEC_RGB565_INV;
         decore(codec_param->dec_handle, DEC_OPT_SETOUT, &codec_param->dec_param, NULL);
         return(0);
      default:
   }
   return(1);
}


int decode_divx( quicktime_t *file, int track,
                 unsigned long inputsize,
                 unsigned char *input, unsigned char **output )
{
   CODEC_PARAM *codec_param;
   DEC_FRAME dec_frame;

   codec_param = (CODEC_PARAM*)((quicktime_codec_t*)file->vtracks[track].codec)->priv;

   dec_frame.bitstream = input;
   dec_frame.length = inputsize;
   dec_frame.stride = 0;

   if ( file->color_model == BC_NONE )
      dec_frame.render_flag = 0;
   else
   {
      dec_frame.render_flag = 1;
      if ( file->color_model != codec_param->dec_cmodel )
         if ( change_dec_cmodel(codec_param,file->color_model) )
            return(1);
   }

   switch ( file->color_model )
   {
      case BC_RGB888:
         dec_frame.bmp = codec_param->buffer;
         if( decore(codec_param->dec_handle, DEC_OPT_FRAME, &dec_frame, NULL) )
            return(1);
         RGBswitch(codec_param,codec_param->buffer,*output);
         break;

      case BC_RGBA8888:
         dec_frame.bmp = codec_param->buffer;
         if( decore(codec_param->dec_handle, DEC_OPT_FRAME, &dec_frame, NULL) )
            return(1);
         RGB32switch(codec_param,codec_param->buffer,*output);
         break;

      default:
         dec_frame.bmp = *output;
         if( decore(codec_param->dec_handle, DEC_OPT_FRAME, &dec_frame, NULL) )
            return(1);
   }

   return(0);
}


int encode_divx( quicktime_t *file, int track,
                 unsigned char **input, unsigned char *output,
                 int *IsAKeyFrame )
{
   CODEC_PARAM *codec_param;
   ENC_RESULT enc_result;
   ENC_FRAME enc_frame;

   codec_param = (CODEC_PARAM*)((quicktime_codec_t*)file->vtracks[track].codec)->priv;

   *IsAKeyFrame = 0;

   enc_frame.bitstream = output;
   enc_frame.length    = 0;
   enc_frame.quant = 0;
   enc_frame.intra = -1;
   enc_frame.mvs = NULL;

   switch ( file->color_model )
   {
      case BC_RGB888:
         RGBswitch(codec_param,*input,codec_param->buffer);
         enc_frame.image = codec_param->buffer;
         enc_frame.colorspace = ENC_CSP_RGB24;
         break;
      case BC_YUV420P:
         enc_frame.image = *input;
         enc_frame.colorspace = ENC_CSP_I420;
         break;
      default:
         return(0);
   }

   enc_frame.quant = 0;
   enc_frame.intra = -1;
   enc_frame.mvs = NULL;

   if ( encore(codec_param->enc_param.handle, ENC_OPT_ENCODE, &enc_frame, &enc_result) )
      return(0);

   if ( enc_result.is_key_frame )
      *IsAKeyFrame = 1;

   return(enc_frame.length);
}


int delete_divx( quicktime_video_map_t *vtrack )
{
   CODEC_PARAM *codec_param;

   codec_param = ((quicktime_codec_t*)vtrack->codec)->priv;

   encore(codec_param->enc_param.handle, ENC_OPT_RELEASE, &codec_param->enc_param, NULL);

   decore(codec_param->dec_handle, DEC_OPT_RELEASE, NULL, NULL);

   free(codec_param->dec_param.buffers.mp4_display_buffers);
   free(codec_param->dec_param.buffers.mp4_edged_for_buffers);
   free(codec_param->dec_param.buffers.mp4_edged_back_buffers);
   free(codec_param->dec_param.buffers.mp4_edged_ref_buffers);
   free(codec_param->dec_param.buffers.mp4_reference);
   free(codec_param->dec_param.buffers.mp4_state);
   free(codec_param->dec_param.buffers.mp4_stream);
   free(codec_param->dec_param.buffers.mp4_tables);

   codec_param->dec_param.buffers.mp4_reference = NULL;

   free(codec_param->buffer);

   free(((quicktime_codec_t*)vtrack->codec)->priv);
   ((quicktime_codec_t*)vtrack->codec)->priv = NULL;

   return(0);
}


int reads_colormodel_divx( quicktime_t *file, int colormodel, int track )
{
   return(    ( colormodel == BC_RGB888 )
           || ( colormodel == BC_RGBA8888 )
           || ( colormodel == BC_BGR888 )
           || ( colormodel == BC_BGRA8888 )
           || ( colormodel == BC_BGR565 )
           || ( colormodel == BC_YUV420P )
           || ( colormodel == BC_NONE )     );
}


int writes_colormodel_divx( quicktime_t *file, int colormodel, int track )
{
   return(    ( colormodel == BC_RGB888 )
           || ( colormodel == BC_YUV420P ) );
}
